/* $Id: eeprom.h,v 1.3 1995/11/25 02:31:38 davem Exp $
 * eeprom.h:  Definitions for the Sun eeprom.
 *
 * Copyright (C) 1995 David S. Miller (davem@caip.rutgers.edu)
 */

/* The EEPROM and the Mostek Mk48t02 use the same IO address space
 * for their registers/data areas.  The IDPROM lives here too.
 */
