/*
 * linux/include/asm-arm/proc-armo/io.h
 */

/* Nothing to do */
#define dma_cache_inv(_start,_size)		do { } while (0)
#define dma_cache_wback(_start,_size)		do { } while (0)
#define dma_cache_wback_inv(_start,_size)	do { } while (0)
